import os
import time
import torch.nn
import numpy as np
from torch import Tensor
from typing import Tuple, List

from SteveRL.train.config import Config
import wandb
class Evaluator:
    def __init__(self,cwd:str,env,args:Config,):
        self.cwd =cwd
        self.env =env
        self.agent_id =args.gpu_id
        self.total_step =0
        self.start_time =time.time()
        self.eval_times =args.eval_times
        self.eval_per_step =args.eval_per_step
        self.eval_step_counter = -self.eval_per_step

        self.save_gap =args.save_gap
        self.save_counter =0
        self.if_keep_save =args.if_keep_save
        self.if_over_write = args.if_over_write
        self.recorder_path = f'{cwd}/recorder.npy'
        self.recorder = []  # total_step, r_avg, r_std, obj_c, ...
        self.max_r = -np.inf
        if_wandb = args.wandb
        print("| Evaluator:"
              "\n| `step`: Number of samples, or total training steps, or running times of `env.step()`."
              "\n| `time`: Time spent from the start of training to this moment."
              "\n| `avgR`: Average value of cumulative rewards, which is the sum of rewards in an episode."
              "\n| `stdR`: Standard dev of cumulative rewards, which is the sum of rewards in an episode."
              "\n| `avgS`: Average of steps in an episode."
              "\n| `objC`: Objective of Critic network. Or call it loss function of critic network."
              "\n| `objA`: Objective of Actor network. It is the average Q value of the critic network."
              f"\n{'#' * 80}\n"
              f"{'ID':<3}{'Step':>8}{'Time':>8} |"
              f"{'avgR':>8}{'stdR':>7}{'avgS':>7}{'stdS':>6} |"
              f"{'expR':>8}{'objC':>7}{'objA':>7}{'etc.':>7}")
        if getattr(env,'num_envs',1) ==1:
            self.get_cumulative_rewards_and_step =self.get_cumulative_rewards_and_step_single_env
        else:
            pass

        if if_wandb:
            self.wandb =True
            pass
        else:
            self.wandb =False
            pass

    def evaluate_and_save(self,actor:torch.nn,steps:int,exp_r:float,logging_tuple:tuple):
        self.total_step+=steps
        if self.total_step < self.eval_step_counter +self.eval_per_step:
            return
        self.eval_step_counter = self.total_step

        rewards_step_ten =self.get_cumulative_rewards_and_step(actor)
        returns =rewards_step_ten[:,0]
        steps =rewards_step_ten[:,1]
        avg_r =returns.mean().item()
        std_r =returns.std().item()
        avg_s =steps.mean().item()
        std_s =steps.std().item()

        train_time =int(time.time() - self.start_time)

        '''record the training information'''
        self.recorder.append((self.total_step,avg_r,std_r,exp_r,*logging_tuple))

        prev_max_r =self.max_r
        self.max_r =max(self.max_r,avg_r)
        print(f"{self.agent_id:<3}{self.total_step:8.2e}{train_time:8.0f} |"
              f"{avg_r:8.2f}{std_r:7.1f}{avg_s:7.0f}{std_s:6.0f} |"
              f"{exp_r:8.2f}{''.join(f'{n:7.2f}' for n in logging_tuple)}")
        # if_save = avg_r > prev_max_r
        if_save =False

        if if_save:
            pass
        if not self.if_keep_save:
            return

        self.save_counter+=1
        actor_path =None
        if if_save:
            if self.if_over_write:
                actor_path = f"{self.cwd}/actor.pt"
            else:
                actor_path = f"{self.cwd}/actor__{self.total_step:012}_{self.max_r:09.3f}.pt"
        elif self.save_counter == self.save_gap:
            self.save_counter = 0
            if self.if_over_write:
                actor_path = f"{self.cwd}/actor.pt"
            else:
                actor_path = f"{self.cwd}/actor__{self.total_step:012}.pt"
        if actor_path:
            torch.save(actor,actor_path)


    def get_cumulative_rewards_and_step_single_env(self,actor)->Tensor:
        '''

        :param actor:
        :return:  一个序列的reward
        '''
        rewards_steps_list =[get_cumulative_rewards_and_steps(self.env,actor)  for  _ in range(self.eval_times)]
        rewards_steps_ten = torch.tensor(rewards_steps_list,dtype=torch.float32)
        return rewards_steps_ten

    def save_training_curve_jpg(self):
        recorder = np.array(self.recorder)

        train_time =int(time.time() - self.start_time)
        total_step =int(self.recorder[-1][0])
        fig_title = f"step_time_maxR_{int(total_step)}_{int(train_time)}_{self.max_r:.3f}"

        draw_learning_curve(recorder,fig_title= fig_title,save_path =f"{self.cwd}/LearningCurve.jpg")
        np.save(self.recorder_path,recorder)









def get_cumulative_rewards_and_steps(env,actor,if_rander:bool =False)->Tuple[float,int]:
    '''用训练好的actor 执行环境看看得到的reward 是多少，
    相当于测试的是test 过程 而不是train过程
    返回的是轨迹的reward和步数
    '''
    max_step =env.max_step
    if_discrete =env.if_discrete
    device =next(actor.parameters()).device

    state =env.reset()
    steps =None
    returns = 0.0

    for steps in range(max_step):
        tensor_state = torch.as_tensor(state,dtype=torch.float32,device=device).unsqueeze(0)
        tensor_action= actor(tensor_state)
        if if_discrete:
            tensor_action = tensor_action.argmax(dim =1)
        action =tensor_action.detach().cpu().numpy()[0]
        state,reward,done,_ =env.step(action)
        returns +=reward

        if if_rander:
            env.render()
            time.sleep(0.02)

        if done:
            break
    else:
        print("| get_rewards_and_step: WARNING. max_step > 12345")
    returns = getattr(env, 'cumulative_returns', returns)
    steps+=1
    return returns,steps

def draw_learning_curve(recorder:np.ndarray =None,fig_title:str ='learning_curve',
                        save_path:str ='learning-curve.jpg'):
    steps =recorder[:,0]
    r_avg =recorder[:,1]
    r_std =recorder[:,2]
    r_exp =recorder[:,3]
    obj_c = recorder[:,4]
    obj_a =recorder[:,5]

    import  matplotlib as mpl
    mpl.use('Agg')
    import  matplotlib.pyplot as  plt
    fig,axs =plt.subplots(2)
    ax00 =axs[0]
    ax00.cla()

    ax01 =axs[0].twinx()
    color01 ='darkcyan'
    ax01.set_ylabel('Explore AvgReward',color =color01)
    ax01.polt(steps,r_exp,color= color01,alpha =0.5)
    ax01.tick_params(axis='y',labelcolor= color01)

    color0 ='lightcoral'
    ax00.set_ylabel('Episode Return',color= color0)
    ax00.plot(steps, r_avg, label='Episode Return', color=color0)
    ax00.fill_between(steps, r_avg - r_std, r_avg + r_std, facecolor=color0, alpha=0.3)
    ax00.grid()
    '''axs[1]'''
    ax10 = axs[1]
    ax10.cla()

    ax11 = axs[1].twinx()
    color11 = 'darkcyan'
    ax11.set_ylabel('objC', color=color11)
    ax11.fill_between(steps, obj_c, facecolor=color11, alpha=0.2, )
    ax11.tick_params(axis='y', labelcolor=color11)

    color10 = 'royalblue'
    ax10.set_xlabel('Total Steps')
    ax10.set_ylabel('objA', color=color10)
    ax10.plot(steps, obj_a, label='objA', color=color10)
    ax10.tick_params(axis='y', labelcolor=color10)
    for plot_i in range(6, recorder.shape[1]):
        other = recorder[:, plot_i]
        ax10.plot(steps, other, label=f'{plot_i}', color='grey', alpha=0.5)
    ax10.legend()
    ax10.grid()

    '''plot save'''
    plt.title(fig_title, y=2.3)
    plt.savefig(save_path)
    plt.close('all')




















